"""In this module we define the ProgressExt widget.

The purpose of the ProgressExt widget is to make the existing `pn.widgets.Progress` easier to use.

The ProgressExt does this by providing

- A combination of a progress value and a progress message
- Easy to use functions for
    - Annotatings
    - Context Management

To enable easy progress reporting from functions and parts of code.

An example use case is

```python
progress = ProgressExt()
run_button = pn.widgets.Button(name="Click me")

@progress.increment(50, "incrementing ...")
def run(event):
    time.sleep(0.5)
run_button.on_click(run)

app = pn.Column(run_button, progress.view)
app.servable()
```

which will show the progress and reset every 2 clicks.
"""
from contextlib import contextmanager

import panel as pn
import param


class ProgressExt(param.Parameterized):
    """The purpose of the ProgressExt widget is to enable easier progress reporting using the
    existing [`pn.widgets.Progress`]\
    (https://panel.pyviz.org/reference/widgets/Progress.html#gallery-progress) widget.

    The ProgressExt widget provides

    - A combination of a progress value and a progress message
    - Easy to use functionality for
        - Function Annotation
        - Context Management

    An example use case is

    ```python
    progress = ProgressExt()
    run_button = pn.widgets.Button(name="Click me")

    @progress.increment(50, "incrementing ...")
    def run(event):
        time.sleep(0.5)
    run_button.on_click(run)

    app = pn.Column(run_button, progress.view)
    app.servable()
    ```

    which will show the progress and reset every 2 clicks.
    """

    value = param.Integer(default=0)
    message = param.String(default="")
    bar_color = param.String(default="info")

    def update(
        self,
        value: int,
        message: str,
    ):
        """Updates the value and message

        Args:
            value (int): A value between 0 and 100
            message (str): A message for the user describing what is happening
        """
        # Please note the order matters as the Widgets updates two times. One for each change
        self.value = value
        self.message = message

    def reset(
        self,
    ):
        """Resets the value and message"""
        # Please note the order matters as the Widgets updates two times. One for each change
        self.message = ""
        self.value = 0

    @param.depends(
        "value",
        "message",
        "bar_color",
    )
    def view(
        self,
    ) -> pn.viewable.Viewable:
        """View the widget

        Returns:
            pn.viewable.Viewable: Add this to your app to see the progress reported
        """
        content = []
        if self.value:
            content.append(
                pn.widgets.Progress(
                    value=self.value,
                    bar_color=self.bar_color,
                    align="center",
                )
            )
        elif self.message:
            content.append(
                pn.widgets.Progress(
                    active=True,
                    bar_color=self.bar_color,
                    align="center",
                )
            )
        if self.value and not self.message:
            content.append(pn.pane.Str(" "))  # Hack: To stop progressbar from jumping vertically
        if self.message:
            content.append(pn.pane.Str(self.message))
        return pn.Row(*content, sizing_mode="stretch_width")

    @contextmanager
    def report(
        self,
        value: int = 0,
        message: str = "",
    ):
        """Report the value and message.

        When the function or code is finished the progress is reset.

        Can be used as context manager or decorator.

        Args:
            value (int, optional): A value between 0 and 100. Default is 0.
            message (str, optional): A message for the user describing what is happening. Default
            is ""

        Yields:
            None: Nothing is yielded
        """
        self.update(
            value,
            message,
        )
        yield
        self.reset()

    @contextmanager
    def increment(
        self,
        value: int,
        message: str,
    ):
        """Increment the value and report the message.

        When the function or code is finished the progress is NOT reset unless self.value >= 100.

        Can be used as context manager or decorator.

        Args:
            value (int): A value between 0 and 100 that will be added to self.value.
            message (str): A message for the user describing what is happening

        Yields:
            None: Nothing is yielded
        """
        value_half = int(value / 2)

        self.value = min(
            self.value + value_half,
            100,
        )
        self.message = message

        yield
        new_value = self.value + value - value_half
        if new_value >= 100:
            self.reset()
        else:
            self.value = new_value



"""The Fast GridTemplate provides a grid layout based on React Grid Layout similar to the
Panel ReactTemplate but in the Fast.design style and enabling the
use of Fast components.
"""
import pathlib

import param
from bokeh.themes import Theme as _BkTheme
from panel.template.theme import DarkTheme, DefaultTheme

from awesome_panel_extensions.frameworks.fast import styles
from awesome_panel_extensions.frameworks.fast.templates.base import GridBasicTemplate


class FastGridTemplate(GridBasicTemplate):
    """
    The FastGridTemplate is build on top of Fast.design and the React Grid Layout.
    """

    _css = pathlib.Path(__file__).parent / "fast_grid_template.css"
    _js = pathlib.Path(__file__).parent.parent.parent / "assets/js/fast_template.js"

    _template = pathlib.Path(__file__).parent / "fast_grid_template.html"

    def _get_theme(self, name: str = "default"):
        if name == "dark":
            return FastGridDarkTheme
        return FastGridDefaultTheme


class FastGridDefaultTheme(DefaultTheme):
    """The Default Theme of the FastGridTemplate"""

    css = param.Filename(default=pathlib.Path(__file__).parent / "default.css")

    _template = FastGridTemplate

    style = param.ClassSelector(class_=styles.FastStyle, default=styles.DEFAULT_STYLE)

    bokeh_theme = param.ClassSelector(
        class_=(_BkTheme, str), default=_BkTheme(json=styles.DEFAULT_BOKEH_THEME)
    )


class FastGridDarkTheme(DarkTheme):
    """The Dark Theme of the FastGridTemplate"""

    css = param.Filename(default=pathlib.Path(__file__).parent / "dark.css")

    _template = FastGridTemplate

    style = param.ClassSelector(class_=styles.FastStyle, default=styles.DARK_STYLE)

    bokeh_theme = param.ClassSelector(
        class_=(_BkTheme, str), default=_BkTheme(json=styles.DARK_BOKEH_THEME)
    )



"""This module provides functionality for defining your site and registrering applications"""
import pathlib
from functools import wraps
from typing import Callable, Dict, Optional

import param
from panel.template.base import BasicTemplate

from awesome_panel_extensions.site.models import Application, User


class Site(param.Parameterized):
    """The Site provides meta data and functionality for registrering application meta data and
    views"""

    applications = param.List(doc="The list of applications to include in the site", constant=True)
    users = param.List(doc="The list of users contributing to the site", constant=True)
    css_path = param.ClassSelector(doc="A path to custom css", class_=pathlib.Path)
    js_path = param.ClassSelector(doc="A path to custom js", class_=pathlib.Path)

    def __init__(self, **params):
        if "applications" not in params:
            params["applications"] = []
        if "users" not in params:
            params["users"] = []

        super().__init__(**params)

    def _get_or_create_user(self, user: str) -> User:
        for item in self.users:  # pylint: disable=not-an-iterable
            if item.uid == user:
                return item
        return User(uid=user, name=user)

    def create_application(  # pylint: disable=too-many-arguments
        self,
        **params,
    ) -> Application:
        """Returns an Application from specified params.

        If you specify the author and owner uid as a string it will automatically be converted into
        a User.

        Returns:
            Application: An application
        """
        for key in ["author", "owner"]:
            value = params.get(key, None)
            if isinstance(value, str):
                params[key] = self._get_or_create_user(value)

        return Application(**params)

    def add(
        self,
        application=None,
    ):
        """Registers your Application and view function
        >>> from awesome_panel_extensions.site.models import User, Application
        >>> from awesome_panel_extensions.site import Site
        >>> site = Site(name="awesome-panel.org")
        >>> marc_skov_madsen = User(
        ...     uid="Marc",
        ...     name="Marc Skov Madsen",
        ...     url="https://datamodelsanalytics.com",
        ...     avatar="https://avatars0.githubusercontent.com/u/42288570",
        ...     resources = {
        ...         "twitter": "https://twitter.com/MarcSkovMadsen",
        ...         "linkedin": "https://www.linkedin.com/in/marcskovmadsen/",
        ...         "github": "https://github.com/MarcSkovMadsen",
        ...     }
        ... )
        >>> site.users.append(marc_skov_madsen)
        >>> application = site.create_application(
        ...     url="home",
        ...     name="Home",
        ...     author="Marc",
        ...     description="The home page",
        ...     description_long="The home page of awesome-panel.org.",
        ...     thumbnail="",
        ...     resources={
        ...         "documentation": "",
        ...         "code": "",
        ...         "gif": "",
        ...         "mp4": "",
        ...         "youtube": "",
        ...     },
        ...     tags=["Site"],
        ... )
        >>> @site.add(application)
        ... def view():
        ...     return pn.pane.Markdown("# Home")
        >>> site.applications
        [Application(name='Home')]
        >>> site.routes
        {'home': <function view at...>}
        """
        # pylint: disable=unsupported-assignment-operation
        if not application.url in [
            app.url for app in self.applications  # pylint: disable=not-an-iterable
        ]:  # pylint: disable=unsupported-membership-test
            self.applications.append(application)

        def inner_function(view):
            @wraps(view)
            def wrapper(*args, **kwargs):
                template = view(*args, **kwargs)
                if (
                    isinstance(template, BasicTemplate)
                    and template.title == template.param.title.default
                ):
                    if not self.name == application.name:
                        template.title = application.name
                    else:
                        template.title = ""
                self.register_post_view(template=template, application=application)
                return template

            application.view = wrapper
            return wrapper

        return inner_function

    # pylint: disable=unused-argument
    def register_post_view(self, template: BasicTemplate, application: Application):
        """Updates the template or application"""

    @property
    def routes(self) -> Dict[str, Callable]:
        """Returns a dictionary with the url as key and the view as the value

        Returns:
            Dict[str, Callable]: [description]
        """
        # pylint: disable=not-an-iterable
        return {app.url: app.view for app in self.applications}

    def get_application(self, name: str) -> Optional[Application]:
        """Returns the application with the specified name

        Args:
            name (str): [description]

        Returns:
            Optional[Application]: [description]
        """
        # pylint: disable=not-an-iterable
        _app = [app for app in self.applications if app.name == name]
        if _app:
            return _app[0]
        return None



"""Base constants like BUTTON_TYPES AND PROGRESS_TYPES"""
_WIDGET_TYPES = [
    "default",
    "primary",
    "success",
    "warning",
    "danger",
]
BUTTON_TYPES = _WIDGET_TYPES
PROGRESS_TYPES = _WIDGET_TYPES



"""The Fast Templates provide easy to use, awesome looking templates for Panel based on the
https://fast.design framework"""
# pylint: disable=line-too-long
from awesome_panel_extensions.frameworks.fast.templates.fast_grid_template.fast_grid_template import (
    FastGridTemplate,
)
from awesome_panel_extensions.frameworks.fast.templates.fast_list_template.fast_list_template import (
    FastListTemplate,
)



"""Functionality for importing the Fast javascript library"""
from panel.pane import HTML

FAST_JS_MODULE = "https://unpkg.com/@microsoft/fast-components@1.13.0"
FAST_JS_SCRIPT = f'<script type="module" src="{FAST_JS_MODULE}"></script>'


def get_fast_js_panel() -> HTML:
    """Returns a panel with the Fast Javascript libary import

    Returns:
        HTML: A Panel to be included in your Fast Application
    """

    return HTML(FAST_JS_SCRIPT, width=0, height=0, sizing_mode="fixed", margin=0)



"""The AwesomeButton extends the `panel.widgets.Button` with support for Icons

    Please note that when the Panel gets extended to support Icons the AwesomeButton will be
    removed.
"""
from typing import Optional

import panel as pn
import param
from panel.widgets.button import Button as _PnButton

from awesome_panel_extensions.models.icon import Icon

_CSS = """
.bk-btn-light .bk-btn-default {
    border-width: 0px;
}

.bk-btn-light .bk-btn-default:hover, .bk-btn-light .bk-btn-default:focus {
            border-width: 1px;
            color: #212529;
            background-color: #e2e6ea;
            border-color: #dae0e5;
}

.bk-btn-light .bk-btn-default:active, .bk-btn-light .bk-btn-default.bk-active  {
            border-width: 1px;
            color: #212529;
            background-color: #dae0e5;
            border-color: #d3d9df;
            box-shadow: none;
}
"""
pn.config.raw_css.append(_CSS)

# pylint: disable=line-too-long
class AwesomeButton(_PnButton):
    """The AwesomeButton extends the Panel Button with support for Icons

    Example:

    >>> icon = Icon(
    ...     name="Github",
    ...     value='<svg xmlns="http://www.w3.org/2000/svg" class="icon icon-github" viewBox="0 0 19 20"><path d="M6.71154 17.0776C2.09615 18.4906 2.09615 14.7226 0.25 14.2517L6.71154 17.0776ZM13.1731 19.9036V16.2581C13.2077 15.8089 13.1482 15.3574 12.9986 14.9335C12.849 14.5096 12.6127 14.123 12.3054 13.7995C15.2038 13.4698 18.25 12.3489 18.25 7.20563C18.2498 5.89046 17.754 4.62572 16.8654 3.67319C17.2862 2.52257 17.2564 1.25074 16.7823 0.121918C16.7823 0.121918 15.6931 -0.207776 13.1731 1.51605C11.0574 0.930913 8.82722 0.930913 6.71154 1.51605C4.19154 -0.207776 3.10231 0.121918 3.10231 0.121918C2.62819 1.25074 2.59844 2.52257 3.01923 3.67319C2.12396 4.63279 1.62771 5.90895 1.63462 7.23389C1.63462 12.3394 4.68077 13.4604 7.57923 13.8278C7.27554 14.148 7.04132 14.5299 6.89182 14.9486C6.74233 15.3674 6.6809 15.8135 6.71154 16.2581V19.9036"></path></svg>',
    ...     fill_color="#E1477E",
    ...     spin_duration=2000,
    ... )
    >>> AwesomeButton(
    ...     name="Click Me",
    ...     icon = icon
    ... )
    AwesomeButton(_bk_icon=Icon(id='...', ...), icon=Icon, name='Click Me'...

    Args:
        icon (Optional[Icon], optional): An Icon. Defaults to None.
        **params: Any parameter the Panel Button supports
    """

    icon = param.ClassSelector(class_=Icon)
    _bk_icon = param.Parameter()
    _rename = {"clicks": None, "name": "label", "value": None, "icon": None, "_bk_icon": "icon"}

    def __init__(self, icon: Optional[Icon] = None, **params):
        if icon:
            params["icon"] = icon
            params["_bk_icon"] = icon._bk_icon
        super().__init__(**params)

    @param.depends("icon", watch=True)
    def _update_bk_icon(self, *_):
        if self.icon:
            self._bk_icon = self.icon._bk_icon  # pylint: disable=protected-access
        else:
            self.icon = None



"""This module implements a View to be used for viewing errors"""
import panel as pn

ERROR_VIEW_CLASS = "designer-error-view"
ERROR_MESSAGE_CLASS = "designer-error-message"
BACKGROUND = "#f8d7da"


class ErrorView(pn.Column):
    """The ErrorView displays the error_message nicely formatted

    Args:
        error_message (str): A stack trace or similar
        params: Parameters normally accepted by panel.Column

    """

    def __init__(self, error_message: str, **params):
        error = error_message
        error = f"""\
# Error

```bash
{error}
```
"""
        if params is None:
            params = {}
        if "css_classes" not in params:
            params["css_classes"] = []
        if not ERROR_VIEW_CLASS in params:
            params["css_classes"].append(ERROR_VIEW_CLASS)
        params["background"] = BACKGROUND
        if not "sizing_mode" in params:
            params["sizing_mode"] = "stretch_width"

        super().__init__(
            pn.pane.Markdown(error, margin=20, css_classes=[ERROR_MESSAGE_CLASS]), **params
        )



"""The FastTextInput enables using the `fast-button` with Bokeh."""
from bokeh.core import properties
from bokeh.models import TextInput as _BkTextInput


class FastTextInput(_BkTextInput):
    """The FastTextInput enables using the `fast-button` with Bokeh."""

    # value  is inherited
    # placeholder is inherited
    # list is not supported
    appearance = properties.String(
        help="""Determines the appearance of the textinput. One of `outline` or `filled`.
        Defaults to outlined""",
    )
    autofocus = properties.Bool(
        help="""The autofocus attribute. Defaults to `False`""",
    )
    type_of_text = properties.String(
        help="""Determines the type of text accepted. One of `email`, `password`, `tel`, `text` or
        `url`. Defaults to text.
        """
    )
    # max_length = properties.Int(
    #     help="""The maximum length of the text string""",
    # )
    min_length = properties.Int(
        help="""The minimum length of the text string""",
    )
    pattern = properties.String(
        help="""A regular expression that the input's value must match in order for the value to
        pass constraint validation"""
    )
    size = properties.Int(
        help="""Valid for email, password, tel, and text input types only. Specifies how much of
        the input is shown""",
    )
    spellcheck = properties.Bool(
        help="""Whether or not the spell check is enabled. Default is False"""
    )
    required = properties.Bool(
        help="""Whether or not the FastTextInput is required. Default is False"""
    )
    readonly = properties.Bool(
        help="""Whether or not the FastTextInput is readonly. Default is False"""
    )



"""The FastTextAreaInput enables using the `fast-textarea-input` with Bokeh."""
from bokeh.core import properties
from bokeh.models import TextAreaInput as _BkTextAreaInput


class FastTextAreaInput(_BkTextAreaInput):
    """The FastTextAreaInput enables using the `fast-textarea-input` with Bokeh."""

    # value  is inherited
    # placeholder is inherited
    # max_length is inherited
    # cols is inherited
    # rows is inherited
    appearance = properties.String(
        help="""Determines the appearance of the textinput. One of `outline` or `filled`.
        Defaults to outlined""",
    )
    autofocus = properties.Bool(
        help="""The autofocus attribute. Defaults to `False`""",
    )
    resize = properties.String(
        help="""The resize attribute. One of
        `None`, `both`, `horizontal` or `vertical`. Defaults to `None`.""",
    )

    min_length = properties.Int(
        help="""The minimum length of the text string""",
    )

    spellcheck = properties.Bool(
        help="""Whether or not the spell check is enabled. Default is False"""
    )
    required = properties.Bool(
        help="""Whether or not the FastTextInput is required. Default is False"""
    )
    readonly = properties.Bool(
        help="""Whether or not the FastTextInput is readonly. Default is False"""
    )



"""Utilities used by awesome-panel.org"""
import pathlib
from typing import Dict, Optional

import panel as pn
import param
from panel.template.base import BasicTemplate

from awesome_panel_extensions._shared.logger import get_logger

# pylint: enable=line-too-long
from awesome_panel_extensions.site.template.template_settings import (
    DEFAULT_TEMPLATE,
    DEFAULT_THEME,
    TemplateSettings,
)

logger = get_logger(__name__)

TEMPLATES: Dict[str, pn.template.BaseTemplate] = {
    "vanilla": pn.template.VanillaTemplate,
    "golden": pn.template.GoldenTemplate,
    "material": pn.template.MaterialTemplate,
    "bootstrap": pn.template.BootstrapTemplate,
    "react": pn.template.ReactTemplate,
    "fast": pn.template.FastListTemplate,
    "fastgrid": pn.template.FastGridTemplate,
}
THEMES = {
    "vanilla": {"default": pn.template.DefaultTheme, "dark": pn.template.DarkTheme},
    "golden": {"default": pn.template.DefaultTheme, "dark": pn.template.DarkTheme},
    "bootstrap": {"default": pn.template.DefaultTheme, "dark": pn.template.DarkTheme},
    "react": {"default": pn.template.DefaultTheme, "dark": pn.template.DarkTheme},
    "material": {
        "default": pn.template.material.MaterialDefaultTheme,
        "dark": pn.template.material.MaterialDarkTheme,
    },
    "fast": {
        "default": pn.template.fast.list.FastDefaultTheme,
        "dark": pn.template.fast.list.FastDarkTheme,
    },
    "fastgrid": {
        "default": pn.template.fast.grid.FastGridDefaultTheme,
        "dark": pn.template.fast.grid.FastDarkTheme,
    },
}

_TEMPLATE_CSS_ID = "/* CUSTOM TEMPLATE CSS */\n"
_TEMPLATE_JS_ID = "// CUSTOM TEMPLATE JS"


class TemplateGenerator(param.Parameterized):
    """The TemplateGenerator can create templates based on the url query arguments"""

    css_path = param.ClassSelector(doc="A path to custom css", class_=pathlib.Path)
    js_path = param.ClassSelector(doc="A path to custom js", class_=pathlib.Path)

    default_template = param.String(doc="The default template to use", default=DEFAULT_TEMPLATE)
    default_theme = param.String(doc="The default theme to use", default=DEFAULT_THEME)
    templates = param.Dict(doc="A dictionary of template name: template class", default=TEMPLATES)
    themes = param.Dict(
        doc="A dictionary of template name: a dictionary of theme name: theme class", default=THEMES
    )

    def _set_template_css(self, template, theme):
        # remove other site css
        pn.config.raw_css = [
            css for css in pn.config.raw_css if not css.startswith(_TEMPLATE_CSS_ID)
        ]
        if not self.css_path:
            return

        files = [
            "all.css",
            f"all_{theme}.css",
            f"{template}.css",
            f"{template}_{theme}.css",
        ]
        text = ""
        for file in files:
            text = ""
            if not file in pn.state.cache:
                file_css_id = f"/* {file} */\n"
                css_file = self.css_path / file
                if css_file.exists():
                    text = _TEMPLATE_CSS_ID + file_css_id + css_file.read_text()
            else:
                text = pn.state.cache[file]
                pn.state.cache.pop(file)
            if text:
                pn.config.raw_css.append(text)

    def _get_template_js(self, template):
        if not self.js_path:
            return ""
        jss = []
        files = [
            f"{template}.js",
        ]
        for file in files:
            if not file in pn.state.cache:
                file_js_id = f" {file} \n"
                text = _TEMPLATE_JS_ID + file_js_id + (self.js_path / file).read_text()
            else:
                text = pn.state.cache[file]
                pn.state.cache.pop(file)
            jss.append(text)
        return "\n".join(jss)

    @staticmethod
    def _get_params(value, class_):
        logger.debug("_get_params %s %s", value, class_)
        if isinstance(value, class_):
            return value
        if isinstance(value, tuple):
            value = [*value]
        elif not isinstance(value, list):
            value = [value]
        # Important to fx. convert @param.depends functions
        value = [pn.panel(item) for item in value]

        if class_ == pn.layout.ListLike:
            return class_(objects=value)
        if class_ == pn.layout.GridSpec:
            logger.debug(
                "grid %s",
                value,
            )
            grid = class_()
            for index, item in enumerate(value):
                print(index, item)
                grid[index, :] = item
            return grid

        return value

    def get_template(  # pylint: disable=too-many-arguments, too-complex
        self,
        template: Optional[str] = None,
        theme: Optional[str] = None,
        **params,
    ) -> BasicTemplate:
        """Returns the specified BasicTemplate

        Args:
            template (str, optional): The name of the template. Defaults to TEMPLATE.
            theme (str, optional): The name of the theme. Defaults to THEME.
            **params: Optional parameters

        Returns:
            BasicTemplate: The specified Template
        """
        logger.info("Getting Template")
        logger.info(pn.state.session_args)
        site_parameters = []
        if not template:
            site_parameters.append("template")
        if not theme:
            site_parameters.append("theme")

        if not template:
            template = pn.state.session_args.get("template", self.default_template)
            if isinstance(template, list):
                template = template[0].decode("utf-8")
                template = template.strip("'").strip('"')
        if not theme:
            theme = pn.state.session_args.get("theme", self.default_theme)
            if isinstance(theme, list):
                theme = theme[0].decode("utf-8")
                theme = theme.strip("'").strip('"')
        # pylint: disable=unsubscriptable-object
        template_class = self.templates.get(str(template), self.templates[self.default_template])
        # pylint: enable=unsubscriptable-object
        # To be fixed with PR https://github.com/holoviz/panel/pull/1694
        if "header" in params:
            params["header"] = self._get_params(
                params["header"], template_class.param.header.class_
            )
        if "main" in params:
            params["main"] = self._get_params(params["main"], template_class.param.main.class_)
        if "sidebar" in params:
            params["sidebar"] = self._get_params(
                params["sidebar"], template_class.param.sidebar.class_
            )
        if "modal" in params:
            params["modal"] = self._get_params(params["modal"], template_class.param.modal.class_)

        self._set_template_css(template, theme)
        # pylint: disable=unsubscriptable-object
        template_instance = template_class(
            theme=self.themes.get(str(template), self.themes[self.default_template]).get(
                str(theme), self.default_theme
            ),
            **params,
        )
        # enable: disable=unsubscriptable-object

        if site_parameters and "fast" not in str(template_class).lower():
            site_settings = TemplateSettings(parameters=site_parameters)
            header = pn.Row(pn.layout.HSpacer(), site_settings.view, sizing_mode="stretch_width")
            template_instance.header.append(header)

        return template_instance


if __name__.startswith("bokeh"):
    FAVICON = (
        "https://raw.githubusercontent.com/MarcSkovMadsen/awesome-panel/"
        "2781d86d4ed141889d633748879a120d7d8e777a/assets/images/favicon.ico"
    )
    template_generator = TemplateGenerator()
    _template = template_generator.get_template(
        title="Test App",
        site="Awesome Panel",
        favicon=FAVICON,
        main_max_width="1100px",
        main=[pn.pane.Markdown("hello world")],
    )
    _template.servable()



"""The SiteConfig model holds your Site configuration and settings"""
from typing import Dict

import param


class SiteConfig(param.Parameterized):
    """Can hold the configuration of a site"""

    site_name = param.String("Awesome Panel Starter")

    gallery_description = param.String("Search and find your app or other resource")

    color_primary = param.String("#A01346")
    color_cycle = param.List(["#A01346"])

    site_prefix = param.String("")

    assets_prefix = param.String()
    avatar_prefix = param.String()
    thumbnails_prefix = param.String()
    mp4_prefix = param.String()
    code_prefix = param.String()

    static_dirs = param.Dict()

    persons = param.Dict()
    applications = param.Dict()

    @property
    def routes(self) -> Dict:
        """Returns the routes to serve in the site

        Returns:
            Dict: The key is the url and the value the servable. Either a .py, .ipynb or a Callable
        """
        # pylint: disable=not-an-iterable, no-member
        return {app.url: app.servable for app in self.applications.values()}



"""Provides the MENU html string appended to all templates

If you need some sort of custom MENU html string feel free to customize this code.

Please note that the default MENU only works in [Fast](https://www.fast.design/) based templates.
"""
from typing import List, Optional

from awesome_panel_extensions.site.models import Application

_ACCENT_COLOR = "#A01346"
_EXPAND = ["Main"]


def _get_collapsed_icon(accent_color: str = _ACCENT_COLOR):
    return f"""
    <svg style="stroke: {accent_color}" width="18" height="18" viewBox="0 0 18 18" fill="none" xmlns="http://www.w3.org/2000/svg" slot="collapsed-icon">
                <path d="M15.2222 1H2.77778C1.79594 1 1 1.79594 1 2.77778V15.2222C1 16.2041 1.79594 17 2.77778 17H15.2222C16.2041 17 17 16.2041 17 15.2222V2.77778C17 1.79594 16.2041 1 15.2222 1Z" stroke-linecap="round" stroke-linejoin="round"></path>
                <path d="M9 5.44446V12.5556" stroke-linecap="round" stroke-linejoin="round"></path>
                <path d="M5.44446 9H12.5556" stroke-linecap="round" stroke-linejoin="round"></path>
            </svg>
    """.replace(
        "\n", ""
    )


def _get_expanded_icon(accent_color: str = _ACCENT_COLOR):
    return f"""
<svg style="stroke: {accent_color}" width="18" height="18" viewBox="0 0 18 18" fill="none" xmlns="http://www.w3.org/2000/svg" slot="expanded-icon">
    <path d="M15.2222 1H2.77778C1.79594 1 1 1.79594 1 2.77778V15.2222C1 16.2041 1.79594 17 2.77778 17H15.2222C16.2041 17 17 16.2041 17 15.2222V2.77778C17 1.79594 16.2041 1 15.2222 1Z" stroke-linecap="round" stroke-linejoin="round"></path>
    <path d="M5.44446 9H12.5556" stroke-linecap="round" stroke-linejoin="round"></path>
</svg>
""".replace(
        "\n", ""
    )


MENU_PRE = """
<fast-accordion id="menu">
"""

MENU_POST = """
</fast-accordion>
"""

MENU_GROUP_PRE = """
    <fast-accordion-item slot="item">
        <h3 slot="heading">Group</h3>{collapsed_icon}{expanded_icon}
        <ul>
"""

MENU_GROUP_POST = """
        </ul>
    </fast-accordion-item>
"""


def _category_sort_key(value):
    if value == "Main":
        return "."
    return value


def _name_sort_key(value):
    if value.name == "Home":
        return "."
    return value.name


def _sort_applications(applications):
    return sorted(applications, key=_name_sort_key)


def _sort_categories(categories):
    return sorted(categories, key=_category_sort_key)


def _group_and_sort(applications):
    result = {}
    for application in applications:
        if application.category not in result:
            result[application.category] = [application]
        else:
            result[application.category].append(application)

    sorted_result = {key: _sort_applications(result[key]) for key in _sort_categories(result)}
    return sorted_result


def to_menu_item(application: Application) -> str:
    """Converts a Application to a Menuitem"""
    return f'<li><a href="{application.url}">{application.name}</a></li>'


def to_menu(
    applications: List[Application],
    accent_color: str = _ACCENT_COLOR,
    expand: Optional[List[str]] = None,
) -> str:
    """Converts a list of Applications to a Menu

    Args:
        applications (List[Application]): The list of Applications
        accent_color (List[Application]): The color of the collapsed and expanded icon
        expand (List[str]): The list of categories to expand. Defaults to ["Main"]

    Returns:
        [str]: The Menu as a HTML string
    """
    if expand is None:
        expand = _EXPAND
    groups = _group_and_sort(applications)
    menu = MENU_PRE
    collapsed_icon = _get_collapsed_icon(accent_color=accent_color)
    expanded_icon = _get_expanded_icon(accent_color=accent_color)
    menu_group_pre = MENU_GROUP_PRE.replace("{collapsed_icon}", collapsed_icon).replace(
        "{expanded_icon}", expanded_icon
    )
    for group, group_applications in groups.items():
        pre = menu_group_pre.replace("Group", group)
        if group in expand:
            pre = pre.replace('slot="item">', 'slot="item" expanded>')
        menu += pre

        menu_items = [to_menu_item(application) for application in group_applications]
        menu += "\n".join(menu_items)

        menu += MENU_GROUP_POST
    menu += MENU_POST
    return menu



"""Exports"""
from . import config
from .fast_anchor import FastAnchor
from .fast_button import FastButton
from .fast_checkbox import FastCheckbox
from .fast_literal_input import FastLiteralInput
from .fast_literalarea_input import FastLiteralAreaInput
from .fast_switch import FastSwitch
from .fast_template import FastTemplate
from .fast_text_input import FastTextInput
from .fast_textarea_input import FastTextAreaInput



"""The BinderLinkButton displayes the Binder badge and if clicked opens the Notebook on Binder
in a new tab"""
import panel as pn
import param

from awesome_panel_extensions.widgets.link_buttons.image_link_button import DerivedImageLinkButton

BINDER_IMAGE_URL = "https://mybinder.org/badge_logo.svg"


class BinderLinkButton(DerivedImageLinkButton):
    """The BinderLinkButton displayes the Binder badge and if clicked opens the Notebook on Binder
    in a new tab"""

    repository = param.String()
    branch = param.String()
    folder = param.String()
    notebook = param.String()

    image_url = param.String(default=BINDER_IMAGE_URL, doc="The url to the image", constant=True)

    width = param.Integer(default=125, bounds=(0, None))

    def __init__(self, **params):
        super().__init__(**params)
        self.style = {}
        self._update_link_url_from_parameters()

    @param.depends(
        "repository", "branch", "folder", "notebook", "height", "width", "sizing_mode", watch=True
    )
    def _update_link_url_from_parameters(self, *_):
        folder = self._html_encode(self.folder)
        notebook = self._html_encode(self.notebook)

        with param.edit_constant(self):
            self.link_url = (
                f"https://mybinder.org/v2/gh/{self.repository}/{self.branch}"
                f"?filepath={folder}%2F{notebook}"
            )


if __name__.startswith("bokeh"):
    button = BinderLinkButton(
        repository="marcskovmadsen/awesome-panel-extensions",
        branch="master",
        folder="examples/panes",
        notebook="WebComponent.ipynb",
    )
    settings_pane = pn.Param(
        button,
        parameters=[
            "repository",
            "branch",
            "folder",
            "notebook",
            "height",
            "width",
            "sizing_mode",
            "margin",
        ],
        background="lightgray",
        sizing_mode="stretch_width",
    )
    app = pn.Column(button, settings_pane, width=500, height=800)
    app.servable()



"""The GridBaseTemplate is a copy of the ReactTemplate"""
import pathlib

import panel as pn
import param
from panel.depends import depends
from panel.layout import Card, GridSpec
from panel.template.base import BasicTemplate as _PnBasicTemplate


# pylint: disable=unused-variable, invalid-name, line-too-long
class TemplateConfig(param.Parameterized):
    """Configuration similar but improved compared to panel.config"""

    css_files = param.Dict(constant=True)

    def __init__(self, **params):
        params["css_files"] = params.get("css_files", {})
        super().__init__(**params)


class BasicTemplate(_PnBasicTemplate):
    """Improvement of Panel BasicTemplate"""

    _theme = "dark"

    sidebar_footer = param.String("", doc="""A HTML string appended to the sidebar""")
    enable_theme_toggle = param.Boolean(
        default=True, doc="If True a switch is to toggle the Theme. Default is True"
    )
    config = param.ClassSelector(class_=TemplateConfig, doc="Similar to panel.config")

    def __init__(self, **params):
        if "theme" not in params:
            if params.get("enable_theme_toggle", self.param.enable_theme_toggle.default):
                self._theme = self._get_theme_from_query_args(default=self._theme)
            params["theme"] = self._get_theme(self._theme)
        else:
            if "dark" in str(params["theme"]).lower():
                self._theme = "dark"
            else:
                self._theme = "default"
        params["config"] = params.get("config", TemplateConfig())
        super().__init__(**params)

        if "header_color" not in params:
            self.header_color = self.theme.style.header_color
        if "header_background" not in params:
            self.header_background = self.theme.style.header_background
        self._update_special_render_vars()

    @staticmethod
    def _get_theme_from_query_args(default: str = "default") -> str:
        theme_arg = pn.state.session_args.get("theme", default)
        if isinstance(theme_arg, list):
            theme_arg = theme_arg[0].decode("utf-8")
            theme_arg = theme_arg.strip("'").strip('"')
        return theme_arg

    def _get_theme(self, name: str = "default"):  # pylint: disable=arguments-differ
        """Should be implemented in child classes"""
        raise NotImplementedError()

    def _update_special_render_vars(self):
        self._render_variables["css_base"] = pathlib.Path(self._css).read_text()
        self._render_variables["css_theme"] = pathlib.Path(self.theme.css).read_text()
        self._render_variables["js"] = pathlib.Path(self._js).read_text()
        self._render_variables["theme"] = self._theme
        self._render_variables["style"] = self.theme.style
        self._render_variables["enable_theme_toggle"] = self.enable_theme_toggle
        self._render_variables["sidebar_footer"] = self.sidebar_footer
        self._render_variables["config"] = self.config


class GridBasicTemplate(BasicTemplate):
    """The GridBaseTemplate is an improvement of the ReactTemplate"""

    compact = param.ObjectSelector(default=None, objects=[None, "vertical", "horizontal", "both"])

    cols = param.Dict(default={"lg": 12, "md": 10, "sm": 6, "xs": 4, "xxs": 2})

    breakpoints = param.Dict(default={"lg": 1200, "md": 996, "sm": 768, "xs": 480, "xxs": 0})

    main = param.ClassSelector(
        class_=GridSpec,
        constant=True,
        doc="""
        A list-like container which populates the main area.""",
    )

    row_height = param.Integer(default=50)

    enable_theme_selection = param.Boolean(
        default=True,
        doc="""If True it will be possible for the
    user to toggle between the 'default' and 'dark' theme. Default""",
    )

    _modifiers = {Card: {"children": {"margin": (20, 20)}, "margin": (10, 5)}}

    _resources = {
        "js": {
            "react": "https://unpkg.com/react@16/umd/react.development.js",
            "react-dom": "https://unpkg.com/react-dom@16/umd/react-dom.development.js",
            "babel": "https://unpkg.com/babel-standalone@latest/babel.min.js",
            "react-grid": "https://cdnjs.cloudflare.com/ajax/libs/react-grid-layout/1.1.1/react-grid-layout.min.js",
        },
        "css": {},
    }

    def __init__(self, **params):
        if "main" not in params:
            params["main"] = GridSpec(ncols=12, mode="override")
        super().__init__(**params)
        self._update_render_vars()

    def _get_theme(self, name: str = "default"):
        """Should be implemented in child classes"""
        raise NotImplementedError()

    def _update_render_items(self, event):
        super()._update_render_items(event)
        if event.obj is not self.main:
            return
        layouts = []
        for i, ((y0, x0, y1, x1), v) in enumerate(self.main.objects.items()):
            if x0 is None:
                x0 = 0
            if x1 is None:
                x1 = 12
            if y0 is None:
                y0 = 0
            if y1 is None:
                y1 = self.main.nrows
            layouts.append({"x": x0, "y": y0, "w": x1 - x0, "h": y1 - y0, "i": str(i + 1)})
        self._render_variables["layouts"] = {"lg": layouts, "md": layouts}

    @depends("cols", "breakpoints", "row_height", "compact", watch=True)
    def _update_render_vars(self):
        self._render_variables["breakpoints"] = self.breakpoints
        self._render_variables["cols"] = self.cols
        self._render_variables["rowHeight"] = self.row_height
        self._render_variables["compact"] = self.compact
